// DRN's Cortex-M0 adaptation of GCC ARM_CM3 port.c from FreeRTOS7.0 distribution
// Merges in M0-specific code from NXP's v5.3 port.c, fixes a couple bugs...
// Uses PSP (Process stack pointer) for tasks, unlike MSP version which used MSP.
// July 2011
// Contact Dave.Nadler@Nadler.com for help or info.

// - optionally sets MSP (main stack pointer) to a separate dedicated stack

// #define DRN_PRINT_DIAGNOSTICS // Requires increased stack size !


/*
    FreeRTOS V7.0.0 - Copyright (C) 2011 Real Time Engineers Ltd.
    

    ***************************************************************************
     *                                                                       *
     *    FreeRTOS tutorial books are available in pdf and paperback.        *
     *    Complete, revised, and edited pdf reference manuals are also       *
     *    available.                                                         *
     *                                                                       *
     *    Purchasing FreeRTOS documentation will not only help you, by       *
     *    ensuring you get running as quickly as possible and with an        *
     *    in-depth knowledge of how to use FreeRTOS, it will also help       *
     *    the FreeRTOS project to continue with its mission of providing     *
     *    professional grade, cross platform, de facto standard solutions    *
     *    for microcontrollers - completely free of charge!                  *
     *                                                                       *
     *    >>> See http://www.FreeRTOS.org/Documentation for details. <<<     *
     *                                                                       *
     *    Thank you for using FreeRTOS, and thank you for your support!      *
     *                                                                       *
    ***************************************************************************


    This file is part of the FreeRTOS distribution.

    FreeRTOS is free software; you can redistribute it and/or modify it under
    the terms of the GNU General Public License (version 2) as published by the
    Free Software Foundation AND MODIFIED BY the FreeRTOS exception.
    >>>NOTE<<< The modification to the GPL is included to allow you to
    distribute a combined work that includes FreeRTOS without being obliged to
    provide the source code for proprietary components outside of the FreeRTOS
    kernel.  FreeRTOS is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
    more details. You should have received a copy of the GNU General Public
    License and the FreeRTOS license exception along with FreeRTOS; if not it
    can be viewed here: http://www.freertos.org/a00114.html and also obtained
    by writing to Richard Barry, contact details for whom are available on the
    FreeRTOS WEB site.

    1 tab == 4 spaces!

    http://www.FreeRTOS.org - Documentation, latest information, license and
    contact details.

    http://www.SafeRTOS.com - A version that is certified for use in safety
    critical systems.

    http://www.OpenRTOS.com - Commercial support, development, porting,
    licensing and training services.
*/

/*-----------------------------------------------------------
 * Implementation of functions defined in portable.h for the ARM CM0 port.
 *----------------------------------------------------------*/

/* Scheduler includes. */
#include "FreeRTOS.h"
#include "task.h"

#ifdef DRN_PRINT_DIAGNOSTICS
  // DRN: CodeRed environment drivers
  #include "driver_config.h" // DRN
  #include "debug_printf.h"  // DRN
#else
  #define debug_printf(...) // DRN: stub out diagnostics below
#endif // DRN_PRINT_DIAGNOSTICS

/* For backward compatibility, ensure configKERNEL_INTERRUPT_PRIORITY is
defined.  The value should also ensure backward compatibility.
FreeRTOS.org versions prior to V4.4.0 did not include this definition. */
#ifndef configKERNEL_INTERRUPT_PRIORITY
    #define configKERNEL_INTERRUPT_PRIORITY 255
#endif

/* Constants required to manipulate the NVIC. */
#define portNVIC_SYSTICK_CTRL       ( ( volatile unsigned long *) 0xe000e010 )
#define portNVIC_SYSTICK_LOAD       ( ( volatile unsigned long *) 0xe000e014 )
#define portNVIC_INT_CTRL           ( ( volatile unsigned long *) 0xe000ed04 )
#define portNVIC_SYSPRI2            ( ( volatile unsigned long *) 0xe000ed20 )
#define portNVIC_SYSTICK_CLK        0x00000004
#define portNVIC_SYSTICK_INT        0x00000002
#define portNVIC_SYSTICK_ENABLE     0x00000001
#define portNVIC_PENDSVSET          0x10000000
#define portNVIC_PENDSV_PRI         ( ( ( unsigned long ) configKERNEL_INTERRUPT_PRIORITY ) << 16 )
#define portNVIC_SYSTICK_PRI        ( ( ( unsigned long ) configKERNEL_INTERRUPT_PRIORITY ) << 24 )

/* Constants required to set up the initial stack. */
#define portINITIAL_XPSR            ( 0x01000000 )

/* The priority used by the kernel is assigned to a variable to make access
from inline assembler easier. */
const unsigned long ulKernelPriority = configKERNEL_INTERRUPT_PRIORITY;

/* Each task maintains its own interrupt status in the critical nesting
variable. */
static unsigned portBASE_TYPE uxCriticalNesting = 0xaaaaaaaa;

/*
 * Setup the timer to generate the tick interrupts.
 */
static void prvSetupTimerInterrupt( void );

/*
 * Exception handlers.
 */
void xPortPendSVHandler( void ) __attribute__ (( naked ));
void xPortSysTickHandler( void );
void vPortSVCHandler( void ) __attribute__ (( naked ));

/*
 * Start first task is a separate function so it can be tested in isolation.
 */
void vPortStartFirstTask( void ) __attribute__ (( naked ));

/*-----------------------------------------------------------*/

/*
 * See header file for description.
 */
portSTACK_TYPE *pxPortInitialiseStack( portSTACK_TYPE *pxTopOfStack, pdTASK_CODE pxCode, void *pvParameters )
{
    /* Simulate the stack frame as it would be created by a context switch
    interrupt. */
    pxTopOfStack--; /* Offset added to account for the way the MCU uses the stack on entry/exit of interrupts. */
    *pxTopOfStack = portINITIAL_XPSR;   /* xPSR */
    pxTopOfStack--;
    *pxTopOfStack = ( portSTACK_TYPE ) pxCode;  /* PC */
    pxTopOfStack--;
    *pxTopOfStack = 0;  /* LR */
    pxTopOfStack -= 5;  /* R12, R3, R2 and R1. */
    *pxTopOfStack = ( portSTACK_TYPE ) pvParameters;    /* R0 */
    pxTopOfStack -= 8;  /* R11, R10, R9, R8, R7, R6, R5 and R4. */

    return pxTopOfStack;
}
/*-----------------------------------------------------------*/

void vPortSVCHandler( void )
{
  __asm volatile (
    " ldr   r3, pxCurrentTCBConst2  \n" /* Restore the context. */
    " ldr r1, [r3]                  \n" /* Use pxCurrentTCBConst to get the pxCurrentTCB address. */
    " ldr r0, [r1]                  \n" /* The first item in pxCurrentTCB is the task top of stack. */
    //==================" ldmia r0!, {r4-r11}               \n"  === Replaced by:
        "                           \n" /* Pop the registers that are not automatically saved on exception entry and the critical nesting count. */
        " ldm r0!, {r4-r7}          \n" /* Pop r8-r11 */
        " mov r8,  r4               \n"
        " mov r9,  r5               \n"
        " mov r10, r6               \n"
        " mov r11, r7               \n"
        " ldm r0!, {r4-r7}          \n" /* Pop r4-r7 */
    //=======
    " msr psp, r0                   \n" /* Restore the task stack pointer. */
    " mov r0, #0                    \n"
    " msr   basepri, r0             \n"
// ??? ommitted in M0:  " orr r14, #0xd  \n" /* DRN: from M3 port - what is this for ??? */
    " bx r14                        \n"
    "                               \n"
    " .align 2                      \n"
    "pxCurrentTCBConst2: .word pxCurrentTCB             \n"
  );
}
/*-----------------------------------------------------------*/

void vPortStartFirstTask( void )
{
  // SVC handler will start first and subsequent tasks using PSP stack,
  // running in "thread mode" as recommended by ARM.

  // Ensure that normal processing runs on PSP ("Process" Stack
  // Pointer, in "thread mode" or "normal mode"). Processor starts
  // out of reset using MSP, so unless the RTL has already switched,
  // we need to switch here.
  // Startup code requires we are already running on PSP...
  debug_printf("CONTROL register=%08x, stack in use is %s\n",
               __get_CONTROL(),
               (__get_CONTROL()&0x2) ? "PSP, OK." : "MSP, need to switch." );
  if( (__get_CONTROL()&0x2) == 0 ) {
    // Ooops, we're still using MSP "Main" Stack Pointer
    #ifdef DRN_PRINT_DIAGNOSTICS
      int answer = 46;
    #endif
    // Set up "process" stack pointer and switch to it
    __set_PSP(__get_MSP()); // copy current stack pointer value into PSP
    __set_CONTROL(0x00000002); // switch to "process" stack pointer PSP
    debug_printf("After stack pointer switch: CONTROL register=%08x, stack in use is %s\n",__get_CONTROL(), (__get_CONTROL()&0x2)?"PSP":"MSP");
    debug_printf("The answer is %d (should be 46)\n",answer); // verify stack pointer is not munged
  }

  // Handler mode (as in processing a hard fault) ALWAYS uses MSP
  // ('Main' stack pointer). For a fault to be trapped properly, MSP
  // must point to a safe stack area.
  __asm volatile(
      // MSP is set to the original reset-stack, thus reusing the stack
      // that main() was using. Because FreeRTOS task stacks were
      // allocated from the heap using using "portMalloc"; anything on
      // the original reset-stack used by "main" will be clobbered.
      // PORT.C also re-uses this reset-stack during context switches.
      " ldr r0, =0x0      \n" /* cm0 has fixed vector table at address 0x0 */
      " ldr r0, [r0]      \n" // fetch first entry in table: reset stack address
      " msr msp, r0       \n" // Set MSP back to the reset stack address
      " cpsie i           \n" /* Globally enable interrupts. */
      " svc 0             \n" /* System call to start first task. */
      " nop               \n"
    );
}
/*-----------------------------------------------------------*/

/*
 * See header file for description.
 */
portBASE_TYPE xPortStartScheduler( void )
{
    /* Make PendSV, CallSV and SysTick the same priroity as the kernel. */
    *(portNVIC_SYSPRI2) |= portNVIC_PENDSV_PRI;
    *(portNVIC_SYSPRI2) |= portNVIC_SYSTICK_PRI;

    /* Start the timer that generates the tick ISR.  Interrupts are disabled
    here already. */
    prvSetupTimerInterrupt();

    /* Initialise the critical nesting count ready for the first task. */
    uxCriticalNesting = 0;

    /* Start the first task. */
    vPortStartFirstTask();

    /* Should not get here! */
    return 0;
}
/*-----------------------------------------------------------*/

void vPortEndScheduler( void )
{
	/* It is unlikely that the CM0 port will require this function as there
    is nothing to return to.  */
}
/*-----------------------------------------------------------*/

void vPortYieldFromISR( void )
{
    /* Set a PendSV to request a context switch. */
    *(portNVIC_INT_CTRL) = portNVIC_PENDSVSET;
}
/*-----------------------------------------------------------*/

void vPortEnterCritical( void )
{
    portDISABLE_INTERRUPTS();
    uxCriticalNesting++;
}
/*-----------------------------------------------------------*/

void vPortExitCritical( void )
{
    uxCriticalNesting--;
    if( uxCriticalNesting == 0 )
    {
        portENABLE_INTERRUPTS();
    }
}
/*-----------------------------------------------------------*/

void xPortPendSVHandler( void )
{
    /* This is a naked function. */

    __asm volatile
    (
        " mrs r0, psp                   \n" // fetch PROCESS stack pointer into R0
        "                               \n"
        " ldr   r3, pxCurrentTCBConst   \n" /* Get the location of the current TCB. */
        " ldr   r2, [r3]                \n"
        "                               \n"
        //====" stmdb r0!, {r4-r11}     \n"  ==== Replaced by:
          "                 \n" /* Save the remaining registers. */
          " sub r0, #4      \n"
          " str r7, [r0]    \n" /* Save r7 */
          " sub r0, #4      \n"
          " str r6, [r0]    \n" /* Save r6 */
          " sub r0, #4      \n"
          " str r5, [r0]    \n" /* Save r5 */
          " sub r0, #4      \n"
          " str r4, [r0]    \n" /* Save r4 */
          "                 \n" 
          " mov r4, r8      \n"
          " mov r5, r9      \n"
          " mov r6, r10     \n"
          " mov r7, r11     \n"
          " sub r0, #4      \n"
          " str r7, [r0]    \n" /* Save r11 */
          " sub r0, #4      \n"
          " str r6, [r0]    \n" /* Save r10 */
          " sub r0, #4      \n"
          " str r5, [r0]    \n" /* Save r9 */
          " sub r0, #4      \n"
          " str r4, [r0]    \n" /* Save r8 */
          "                 \n"
        //=================
        " str r0, [r2]      \n" /* Save the new top of stack into the first member of the TCB. */
        "                   \n"
        //====" stmdb sp!, {r3, r14} \n" ==== Replaced by:
          " ldr r0, =0x0    \n" /* cm0 has fixed vector table at address 0x0 */
          " ldr r0, [r0]    \n" // fetch first entry in table: reset stack address
          " mov r4, r14     \n" 
          " str r4, [r0]    \n" /* Save r14 */
          " msr psp, r0     \n" // Set the PSP back to the reset stack address
          "                 \n"
        //=================
        " mov r0, %0            \n"
        " msr basepri, r0       \n"
        " bl vTaskSwitchContext \n"
        " mov r0, #0            \n"
        " msr basepri, r0       \n"
        //====" ldmia sp!, {r3, r14} \n" ==== replaced by...
          "                 \n"
          " ldr r0, =0x0    \n" /* cm0 has fixed vector table at address 0x0 */
          " ldr r0, [r0]    \n" // fetch first entry in table: reset stack address
          " ldr r4, [r0]    \n" /* Restore r14 to reset stack address */
          " mov r14,r4      \n"
       //=================
       "                            \n" /* Restore the context, including the critical nesting count. */
       " ldr r3, pxCurrentTCBConst  \n" /* From NXP M0 port: Get the location of the current TCB. */
       " ldr r1, [r3]               \n"
       " ldr r0, [r1]               \n" /* The first item in pxCurrentTCB is the task top of stack. */
       //====" ldmia r0!, {r4-r11}                  \n"  ==== Replaced by:
         "                  \n" /* Pop the registers. */
         " ldm r0!, {r4-r7} \n" /* Pop r8-r11 */
         " mov r8,  r4      \n"
         " mov r9,  r5      \n"
         " mov r10, r6      \n"
         " mov r11, r7      \n"
         " ldm r0!, {r4-r7} \n" /* Pop r4-r7 */
      //=================
      " msr psp, r0         \n" // restore stack pointer for this task
      " bx r14              \n" // and return
      "                                         \n"
      " .align 2                                \n"
      "pxCurrentTCBConst: .word pxCurrentTCB    \n"
      "                                         \n"
      ::"i"(configMAX_SYSCALL_INTERRUPT_PRIORITY)
    );
}
/*-----------------------------------------------------------*/

void xPortSysTickHandler( void )
{
unsigned long ulDummy;

    /* If using preemption, also force a context switch. */
    #if configUSE_PREEMPTION == 1
        *(portNVIC_INT_CTRL) = portNVIC_PENDSVSET;
    #endif

    ulDummy = portSET_INTERRUPT_MASK_FROM_ISR();
    {
        vTaskIncrementTick();
    }
    portCLEAR_INTERRUPT_MASK_FROM_ISR( ulDummy );
}
/*-----------------------------------------------------------*/

/*
 * Setup the systick timer to generate the tick interrupts at the required
 * frequency.
 */
void prvSetupTimerInterrupt( void )
{
    /* Configure SysTick to interrupt at the requested rate. */
    *(portNVIC_SYSTICK_LOAD) = ( configCPU_CLOCK_HZ / configTICK_RATE_HZ ) - 1UL;
    *(portNVIC_SYSTICK_CTRL) = portNVIC_SYSTICK_CLK | portNVIC_SYSTICK_INT | portNVIC_SYSTICK_ENABLE;
}
/*-----------------------------------------------------------*/
